@@ -12,9 +12,10 @@ Functions for paths and files.
 #
 #  Copyright © 2018-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
 #
-#  Parts of the docstrings and the PathPlus class based on the Python 3.8.2 Documentation
+#  Parts of the docstrings, the PathPlus class and the DirComparator class
+#  based on Python and its Documentation
 #  Licensed under the Python Software Foundation License Version 2.
-#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
+#  Copyright © 2001-2021 Python Software Foundation. All rights reserved.
 #  Copyright © 2000 BeOpen.com. All rights reserved.
 #  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
 #  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
@@ -30,7 +31,7 @@ Functions for paths and files.
 #
 #  This program is distributed in the hope that it will be useful,
 #  but WITHOUT ANY WARRANTY; without even the implied warranty of
-#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 #  GNU Lesser General Public License for more details.
 #
 #  You should have received a copy of the GNU Lesser General Public License
@@ -41,6 +42,7 @@ Functions for paths and files.
 
 # stdlib
 import contextlib
+import filecmp
 import fnmatch
 import gzip
 import json
@@ -49,10 +51,26 @@ import pathlib
 import shutil
 import stat
 import sys
-from collections import deque
-from typing import IO, Any, Callable, Iterable, Iterator, List, Optional, TypeVar, Union
+import tempfile
+from collections import defaultdict, deque
+from operator import methodcaller
+from typing import (
+		IO,
+		Any,
+		Callable,
+		ContextManager,
+		Dict,
+		Iterable,
+		Iterator,
+		List,
+		Optional,
+		Sequence,
+		TypeVar,
+		Union
+		)
 
 # this package
+from domdf_python_tools.compat import nullcontext
 from domdf_python_tools.typing import JsonLibrary, PathLike
 
 __all__ = [
@@ -75,6 +93,10 @@ __all__ = [
 		"traverse_to_file",
 		"matchglob",
 		"unwanted_dirs",
+		"TemporaryPathPlus",
+		"sort_paths",
+		"DirComparator",
+		"compare_dirs",
 		]
 
 newline_default = object()
@@ -325,12 +347,12 @@ class PathPlus(pathlib.Path):
 	"""
 	Subclass of :class:`pathlib.Path` with additional methods and a default encoding of UTF-8.
 
-	Path represents a filesystem path but unlike :class:`~.PurePath`, also offers
+	Path represents a filesystem path but unlike :class:`pathlib.PurePath`, also offers
 	methods to do system calls on path objects.
 	Depending on your system, instantiating a :class:`~.PathPlus` will return
 	either a :class:`~.PosixPathPlus` or a :class:`~.WindowsPathPlus`. object.
-	You can also instantiate a :class:`PosixPath` or :class:`WindowsPath` directly,
-	but cannot instantiate a :class:`WindowsPath` on a POSIX system or vice versa.
+	You can also instantiate a :class:`~.PosixPathPlus` or :class:`WindowsPath` directly,
+	but cannot instantiate a :class:`~.WindowsPathPlus` on a POSIX system or vice versa.
 
 	.. versionadded:: 0.3.8
 	.. versionchanged:: 0.5.1  Defaults to Unix line endings (``LF``) on all platforms.
@@ -449,6 +471,8 @@ class PathPlus(pathlib.Path):
 			data: Iterable[str],
 			encoding: Optional[str] = "UTF-8",
 			errors: Optional[str] = None,
+			*,
+			trailing_whitespace: bool = False
 			) -> None:
 		"""
 		Write the given list of lines to the file without trailing whitespace.
@@ -458,9 +482,19 @@ class PathPlus(pathlib.Path):
 		:param data:
 		:param encoding: The encoding to write to the file in.
 		:param errors:
+		:param trailing_whitespace: If :py:obj:`True` trailing whitespace is preserved.
+
+		.. versionchanged:: 2.4.0  Added the ``trailing_whitespace`` option.
 		"""
 
-		return self.write_clean('\n'.join(data), encoding=encoding, errors=errors)
+		if trailing_whitespace:
+			data = list(data)
+			if data[-1].strip():
+				data.append('')
+
+			self.write_text('\n'.join(data), encoding=encoding, errors=errors)
+		else:
+			self.write_clean('\n'.join(data), encoding=encoding, errors=errors)
 
 	def read_text(
 			self,
@@ -629,7 +663,7 @@ class PathPlus(pathlib.Path):
 				**kwargs,
 				)
 
-	if sys.version_info < (3, 7):  # pragma: no cover (<py37)
+	if sys.version_info < (3, 10):  # pragma: no cover (<py310)
 
 		def is_mount(self) -> bool:
 			"""
@@ -643,9 +677,9 @@ class PathPlus(pathlib.Path):
 			if not self.exists() or not self.is_dir():
 				return False
 
-			parent = pathlib.Path(self.parent)
+			# https://github.com/python/cpython/pull/18839/files
 			try:
-				parent_dev = parent.stat().st_dev
+				parent_dev = self.parent.stat().st_dev
 			except OSError:
 				return False
 
@@ -653,7 +687,7 @@ class PathPlus(pathlib.Path):
 			if dev != parent_dev:
 				return True
 			ino = self.stat().st_ino
-			parent_ino = parent.stat().st_ino
+			parent_ino = self.parent.stat().st_ino
 			return ino == parent_ino
 
 	if sys.version_info < (3, 8):  # pragma: no cover (<py38)
@@ -759,6 +793,7 @@ class PathPlus(pathlib.Path):
 			self: _PP,
 			exclude_dirs: Optional[Iterable[str]] = unwanted_dirs,
 			match: Optional[str] = None,
+			matchcase: bool = True,
 			) -> Iterator[_PP]:
 		"""
 		Returns an iterator over all children (files and directories) of the current path object.
@@ -769,29 +804,34 @@ class PathPlus(pathlib.Path):
 			together with their children.
 		:param match: A pattern to match filenames against.
 			The pattern should be in the format taken by :func:`~.matchglob`.
+			:param matchcase: Whether the filename's case should match the pattern.
+
+		:rtype:
+
+		.. versionchanged:: 2.5.0  Added the ``matchcase`` option.
 		"""
 
-		if not self.is_dir():
+		if not self.abspath().is_dir():
 			return
 
 		if exclude_dirs is None:
 			exclude_dirs = ()
 
-		if match and not os.path.isabs(match):
+		if match and not os.path.isabs(match) and self.is_absolute():
 			match = (self / match).as_posix()
 
 		file: _PP
-		for file in self.iterdir():  # type: ignore
+		for file in self.iterdir():
 			parts = file.parts
 			if any(d in parts for d in exclude_dirs):
 				continue
 
+			if match is None or (match is not None and matchglob(file, match, matchcase)):
+				yield file
+
 			if file.is_dir():
 				yield from file.iterchildren(exclude_dirs, match)
 
-			if match is None or (match is not None and matchglob(file, match)):
-				yield file
-
 
 class PosixPathPlus(PathPlus, pathlib.PurePosixPath):
 	"""
@@ -863,7 +903,7 @@ def traverse_to_file(base_directory: _P, *filename: PathLike, height: int = -1)
 	raise FileNotFoundError(f"'{filename[0]!s}' not found in {base_directory}")
 
 
-def matchglob(filename: PathLike, pattern: str):
+def matchglob(filename: PathLike, pattern: str, matchcase: bool = True):
 	"""
 	Given a filename and a glob pattern, return whether the filename matches the glob.
 
@@ -873,12 +913,19 @@ def matchglob(filename: PathLike, pattern: str):
 	:param pattern: A pattern structured like a filesystem path, where each element consists of the glob syntax.
 		Each element is matched by :mod:`fnmatch`.
 		The special element ``**`` matches zero or more files or directories.
+	:param matchcase: Whether the filename's case should match the pattern.
+
+	:rtype:
 
 	.. seealso::
 
 		:wikipedia:`Glob (programming)#Syntax` on Wikipedia
+
+	.. versionchanged:: 2.5.0  Added the ``matchcase`` option.
 	"""
 
+	match_func = fnmatch.fnmatchcase if matchcase else fnmatch.fnmatch
+
 	filename = PathPlus(filename)
 
 	pattern_parts = deque(pathlib.PurePath(pattern).parts)
@@ -912,16 +959,188 @@ def matchglob(filename: PathLike, pattern: str):
 				# Filename must match everything after **
 				return False
 
-			if fnmatch.fnmatchcase(filename_part, pattern_part):
+			if match_func(filename_part, pattern_part):
 				continue
 			else:
-				while not fnmatch.fnmatchcase(filename_part, pattern_part):
+				while not match_func(filename_part, pattern_part):
 					if not filename_parts:
 						return False
 
 					filename_part = filename_parts.popleft()
 
-		elif fnmatch.fnmatchcase(filename_part, pattern_part):
+		elif match_func(filename_part, pattern_part):
 			continue
 		else:
 			return False
+
+
+class TemporaryPathPlus(tempfile.TemporaryDirectory):
+	"""
+	Securely creates a temporary directory using the same rules as :func:`tempfile.mkdtemp`.
+	The resulting object can be used as a context manager.
+	On completion of the context or destruction of the object
+	the newly created temporary directory and all its contents are removed from the filesystem.
+
+	Unlike :func:`tempfile.TemporaryDirectory` this class is based around a :class:`~.PathPlus` object.
+
+	.. versionadded:: 2.4.0
+	"""
+
+	name: PathPlus  # type: ignore
+	"""
+	The temporary directory itself.
+
+	This will be assigned to the target of the :keyword:`as` clause if the :class:`~.TemporaryPathPlus`
+	is used as a context manager.
+	"""
+
+	def __init__(
+			self,
+			suffix: Optional[str] = None,
+			prefix: Optional[str] = None,
+			dir: Optional[PathLike] = None,  # noqa: A002  # pylint: disable=redefined-builtin
+			) -> None:
+
+		super().__init__(suffix, prefix, dir)
+		self.name = PathPlus(self.name)
+
+	def cleanup(self) -> None:
+		"""
+		Cleanup the temporary directory by removing it and its contents.
+
+		If the :class:`~.TemporaryPathPlus` is used as a context manager
+		this is called when leaving the :keyword:`with` block.
+		"""
+
+		context: ContextManager
+
+		if sys.platform == "win32":  # pragma: no cover (!Windows)
+			context = contextlib.suppress(PermissionError, NotADirectoryError)
+		else:  # pragma: no cover (Windows)
+			context = nullcontext()
+
+		with context:
+			super().cleanup()
+
+	def __enter__(self) -> PathPlus:
+		return self.name
+
+
+def sort_paths(*paths: PathLike) -> List[PathPlus]:
+	"""
+	Sort the ``paths`` by directory, then by file.
+
+	:param paths:
+
+	.. versionadded:: 2.6.0
+	"""
+
+	directories: Dict[str, List[PathPlus]] = defaultdict(list)
+	local_contents: List[PathPlus] = []
+	files: List[PathPlus] = []
+
+	for obj in [PathPlus(path) for path in paths]:
+		if len(obj.parts) > 1:
+			key = obj.parts[0]
+			directories[key].append(obj)
+		else:
+			local_contents.append(obj)
+
+	# sort directories
+	directories = {directory: directories[directory] for directory in sorted(directories.keys())}
+
+	for directory, contents in directories.items():
+		contents = [path.relative_to(directory) for path in contents]
+		files.extend(PathPlus(directory) / path for path in sort_paths(*contents))
+
+	return files + sorted(local_contents, key=methodcaller("as_posix"))
+
+
+class DirComparator(filecmp.dircmp):
+	r"""
+	Compare the content of ``a`` and ``a``.
+
+	In contrast with :class:`filecmp.dircmp`, this
+	subclass compares the content of files with the same path.
+
+	.. versionadded:: 2.7.0
+
+	:param a: The "left" directory to compare.
+	:param b: The "right" directory to compare.
+	:param ignore: A list of names to ignore.
+	:default ignore: :py:obj:`filecmp.DEFAULT_IGNORES`
+	:param hide: A list of names to hide.
+	:default hide: ``[`` :py:obj:`os.curdir`, :py:obj:`os.pardir` ``]``
+	"""
+
+	# From https://stackoverflow.com/a/24860799, public domain.
+	# Thanks Philippe
+
+	def __init__(
+			self,
+			a: PathLike,
+			b: PathLike,
+			ignore: Optional[Sequence[str]] = None,
+			hide: Optional[Sequence[str]] = None,
+			):
+		super().__init__(a, b, ignore=ignore, hide=hide)
+
+	def phase3(self) -> None:  # noqa: D102
+		# Find out differences between common files.
+		# Ensure we are using content comparison with shallow=False.
+
+		fcomp = filecmp.cmpfiles(self.left, self.right, self.common_files, shallow=False)
+		self.same_files, self.diff_files, self.funny_files = fcomp
+
+	def phase4(self) -> None:  # noqa: D102
+		# Find out differences between common subdirectories
+
+		# From https://github.com/python/cpython/pull/23424
+
+		self.subdirs = {}
+
+		for x in self.common_dirs:
+			a_x = os.path.join(self.left, x)
+			b_x = os.path.join(self.right, x)
+			self.subdirs[x] = self.__class__(a_x, b_x, self.ignore, self.hide)
+
+	_methodmap = {
+			"subdirs": phase4,
+			"same_files": phase3,
+			"diff_files": phase3,
+			"funny_files": phase3,
+			"common_dirs": filecmp.dircmp.phase2,
+			"common_files": filecmp.dircmp.phase2,
+			"common_funny": filecmp.dircmp.phase2,
+			"common": filecmp.dircmp.phase1,
+			"left_only": filecmp.dircmp.phase1,
+			"right_only": filecmp.dircmp.phase1,
+			"left_list": filecmp.dircmp.phase0,
+			"right_list": filecmp.dircmp.phase0
+			}
+
+	methodmap = _methodmap  # type: ignore
+
+
+def compare_dirs(a: PathLike, b: PathLike):
+	"""
+	Compare the content of two directory trees.
+
+	.. versionadded:: 2.7.0
+
+	:param a: The "left" directory to compare.
+	:param b: The "right" directory to compare.
+
+	:returns: :py:obj:`False` if they differ, :py:obj:`True` is they are the same.
+	"""
+
+	compared = DirComparator(a, b)
+
+	if compared.left_only or compared.right_only or compared.diff_files or compared.funny_files:
+		return False
+
+	for subdir in compared.common_dirs:
+		if not compare_dirs(os.path.join(a, subdir), os.path.join(b, subdir)):
+			return False
+
+	return True
